/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the OX Software GmbH group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016-2020 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.util.custom.config;

import java.net.URI;
import java.net.URISyntaxException;
import java.nio.charset.Charset;
import java.util.Iterator;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;
import javax.annotation.CheckReturnValue;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import javax.mail.internet.idn.IDNA;
import com.google.common.base.CharMatcher;
import com.google.common.base.Function;
import com.google.common.base.Optional;
import com.google.common.base.Splitter;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Iterators;
import com.openexchange.config.cascade.ConfigView;
import com.openexchange.config.cascade.ConfigViewFactory;
import com.openexchange.configuration.ConfigurationExceptionCodes;
import com.openexchange.exception.OXException;
import com.openexchange.session.Session;


/**
 * {@link ConfigViewReader}
 *
 * @author <a href="mailto:pascal.bleser@open-xchange.com">Pascal Bleser</a>
 * @since v1.2.0
 */
@ParametersAreNonnullByDefault
public class ConfigViewReader {

    public static ConfigViewReader ofUser(final @Nullable ConfigViewFactory configViewFactory, int userId, int contextId) throws OXException {
        return new ConfigViewReader(null, configViewFactory != null ? configViewFactory.getView(userId, contextId) : null);
    }

    public static ConfigViewReader ofUser(final String prefix, final @Nullable ConfigViewFactory configViewFactory, int userId, int contextId) throws OXException {
        return new ConfigViewReader(prefix, configViewFactory != null ? configViewFactory.getView(userId, contextId) : null);
    }

    public static ConfigViewReader ofUser(final String prefix, final @Nullable ConfigViewFactory configViewFactory, final Session session) throws OXException {
        return new ConfigViewReader(prefix, configViewFactory != null ? configViewFactory.getView(session.getUserId(), session.getContextId()) : null);
    }
    
    public static ConfigViewReader of(final @Nullable ConfigView configView) {
        return new ConfigViewReader(null, configView);
    }

    public static ConfigViewReader of(final String prefix, final @Nullable ConfigView configView) {
        return new ConfigViewReader(prefix, configView);
    }
    
    private final @Nullable ConfigView configView;
    private final @Nullable String prefix;
    
    private ConfigViewReader(final @Nullable String prefix, final @Nullable ConfigView configView) {
        this.configView = configView;
        if (prefix == null || prefix.isEmpty()) {
            this.prefix = null;
        } else {
            this.prefix = prefix;
        }
    }
    
    private final String n(final String name) {
        return prefix(prefix, name);
    }
    
    private static final class ConfigPropertyPrefixFunction implements Function<String, String> {
        private final @Nullable String prefix;
        public ConfigPropertyPrefixFunction(final @Nullable String prefix) {
            this.prefix = prefix;
        }
        @Override
        @Nullable
        public String apply(@Nullable String input) {
            return input != null ? prefix(prefix, input) : null;
        }
        @Override
        public boolean equals(@Nullable Object obj) {
            return super.equals(obj);
        }
    }
    
    public static final Function<String, String> toPrefixFunction(final @Nullable String prefix) {
        return new ConfigPropertyPrefixFunction(prefix);
    }
    
    @SuppressWarnings("null")
    public static final String prefix(final @Nullable String prefix, final String name) {
        if (prefix == null) {
            return name;
        }
        if (name.startsWith(".") && prefix.endsWith(".")) {
            return new StringBuilder(prefix.length() + name.length() - 1)
                .append(prefix)
                .append(name.substring(1))
                .toString();
        }
        if (! (name.startsWith(".") || prefix.endsWith("."))) {
            return new StringBuilder(prefix.length() + name.length() + 1)
                .append(prefix)
                .append(".")
                .append(name)
                .toString();
        }
        return new StringBuilder(prefix.length() + name.length())
            .append(prefix)
            .append(name)
            .toString();
    }
    
    public final class MandatoryValueReader {
        @CheckReturnValue
        public String str(final String name) throws OXException {
            final String n = n(name);
            final ConfigView cv = configView;
            if (cv == null) {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(ConfigView.class.getSimpleName() + " is null");
            }
            String str = cv.opt(n, String.class, null);
            if (str == null) {
                throw ConfigurationExceptionCodes.PROPERTY_MISSING.create(n);
            }
            str = str.replaceAll("^\\s*", "").replaceAll("\\s*$", "");
            if (str.isEmpty()) {
                throw ConfigurationExceptionCodes.PROPERTY_MISSING.create(n);
            }
            return str;
        }
        
        @CheckReturnValue
        public String str(final String name, final String defaultValue) throws OXException {
            final String n = n(name);
            final ConfigView cv = configView;
            if (cv == null) {
                return defaultValue;
            }
            String str = cv.opt(n, String.class, null);
            if (str == null) {
                return defaultValue;
            }
            str = str.replaceAll("^\\s*", "").replaceAll("\\s*$", "");
            if (str.isEmpty()) {
                return defaultValue;
            }
            return str;
        }
        
        @CheckReturnValue
        public long longNum(final String name, final long defaultValue) throws OXException {
            final Optional<String> str = optional.str(name);
            if (! str.isPresent()) {
                return defaultValue;
            }
            try {
                return Long.parseLong(str.get());
            } catch (final NumberFormatException e) {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("%s is not a numeric value", n(name)));
            }
        }

        @CheckReturnValue
        public long longNum(final String name) throws OXException {
            final String str = mandatory.str(name);
            try {
                return Long.parseLong(str);
            } catch (final NumberFormatException e) {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("%s is not a numeric value", n(name)));
            }
        }
        
        @CheckReturnValue
        public int intNum(final String name, final int defaultValue) throws OXException {
            final Optional<String> str = optional.str(name);
            if (! str.isPresent()) {
                return defaultValue;
            }
            try {
                return Integer.parseInt(str.get());
            } catch (final NumberFormatException e) {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("%s is not a numeric value", n(name)));
            }
        }
        @CheckReturnValue
        public int intNum(final String name) throws OXException {
            final String str = mandatory.str(name);
            try {
                return Integer.parseInt(str);
            } catch (final NumberFormatException e) {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("%s is not a numeric value", n(name)));
            }
        }
        @CheckReturnValue
        public int positiveIntNum(final String name) throws OXException {
            final int i = mandatory.intNum(name);
            if (i < 0) {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("%s must be a positive numeric value", n(name)));
            } else {
                return i;
            }
        }
        
        @CheckReturnValue
        public boolean bool(final String name, final boolean defaultValue) throws OXException {
            final Optional<String> o = optional.str(name);
            if (! o.isPresent()) {
                return defaultValue;
            }
            final String str = o.get();
            if (str.equalsIgnoreCase("true")) {
                return true;
            }
            if (str.equalsIgnoreCase("false")) {
                return false;
            }
            throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("%s is not a boolean (true/false)", n(name)));
        }

        @CheckReturnValue
        public boolean bool(final String name) throws OXException {
            final String str = mandatory.str(name);
            if (str.equalsIgnoreCase("true")) {
                return true;
            }
            if (str.equalsIgnoreCase("false")) {
                return false;
            }
            throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("%s is not a boolean (true/false)", n(name)));
        }
        
        @CheckReturnValue
        public String hostname(final String name, final String defaultValue) throws OXException {
            final String str = IDNA.toASCII(mandatory.str(name, defaultValue));
            if (str != null) {
                return str;
            } else {
                return defaultValue;
            }
        }
        @CheckReturnValue
        public String hostname(final String name) throws OXException {
            final String str = IDNA.toASCII(mandatory.str(name));
            if (str != null) {
                return str;
            } else {
                throw ConfigurationExceptionCodes.PROPERTY_MISSING.create(n(name));
            }
        }
        
        @CheckReturnValue
        public int port(final String name) throws OXException {
            final int value = mandatory.intNum(name);
            if (value <= 0 || value > 65535) {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("%s does not contain a valid port number", n(name)));
            } else {
                return value;
            }
        }
        @CheckReturnValue
        public int port(final String name, final int defaultValue) throws OXException {
            final Optional<Integer> o = optional.intNum(name);
            if (! o.isPresent()) {
                return defaultValue;
            }
            final int value = o.get();
            if (value <= 0 || value > 65535) {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("%s does not contain a valid port number", n(name)));
            } else {
                return value;
            }
        }
        
        @CheckReturnValue
        public Charset charset(final String name, final Charset defaultValue) throws OXException {
            final Optional<String> o = optional.str(name);
            if (! o.isPresent()) {
                return defaultValue;
            } else {
                try {
                    return Optional.fromNullable(Charset.forName(o.get())).or(defaultValue);
                } catch (final Exception e) {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("%s does not contain a valid charset: %s",
                        n(name), e.getMessage()));
                }
            }
        }
        @CheckReturnValue
        public Charset charset(final String name) throws OXException {
            final String str = mandatory.str(name);
            final @Nullable Charset result;
            try {
                result = Charset.forName(str);
            } catch (final Exception e) {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("%s does not contain a valid charset: %s",
                    n(name), e.getMessage()));
            }
            if (result != null) {
                return result;
            } else {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("%s does not contain a valid charset", n(name)));
            }
        }
        
        @CheckReturnValue
        public ImmutableList<String> strList(final String name, final CharMatcher split) throws OXException {
            final String str = mandatory.str(name);
            final ImmutableList<String> list = ImmutableList.copyOf(Splitter.on(split).omitEmptyStrings().split(str));
            if (list.isEmpty()) {
                throw ConfigurationExceptionCodes.PROPERTY_MISSING.create(n(name));
            } else {
                return list;
            }
        }
        @CheckReturnValue
        public ImmutableList<String> strList(final String name) throws OXException {
            return mandatory.strList(name, DEFAULT_SPLIT_PATTERN);
        }
        
        @CheckReturnValue
        public ImmutableSet<String> strSet(final String name, final CharMatcher split) throws OXException {
            final String str = mandatory.str(name);
            final ImmutableSet<String> set = ImmutableSet.copyOf(Splitter.on(split).omitEmptyStrings().split(str));
            if (set.isEmpty()) {
                throw ConfigurationExceptionCodes.PROPERTY_MISSING.create(n(name));
            } else {
                return set;
            }
        }
        @CheckReturnValue
        public ImmutableSet<String> strSet(final String name) throws OXException {
            return mandatory.strSet(name, DEFAULT_SPLIT_PATTERN);
        }
        
        @CheckReturnValue
        public URI uri(String name) throws OXException {
            final String str = mandatory.str(name);
            try {
                return new URI(str);
            } catch (final URISyntaxException e) {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("the value of %s is not a valid URI", n(name)));
            }
        }
        
        @CheckReturnValue
        public URI uri(String name, URI defaultValue) throws OXException {
            final Optional<String> str = optional.str(name);
            if (str.isPresent()) {
                try {
                    return new URI(str.get());
                } catch (final URISyntaxException e) {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("the value of %s is not a valid URI", n(name)));
                }
            } else {
                return defaultValue;
            }
        }
        
        @CheckReturnValue
        public URI uri(String name, String defaultValue) throws OXException {
            final Optional<String> str = optional.str(name);
            if (str.isPresent()) {
                try {
                    return new URI(str.get());
                } catch (final URISyntaxException e) {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("the value of %s is not a valid URI", n(name)));
                }
            } else {
                try {
                    return new URI(defaultValue);
                } catch (final URISyntaxException e) {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("the default value of %s is not a valid URI", n(name)));
                }
            }
        }
        
        @CheckReturnValue
        public Pattern pattern(String name) throws OXException {
            final String str = mandatory.str(name);
            final Pattern result;
            try {
                result = Pattern.compile(str);
            } catch (PatternSyntaxException e) {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("the value of %s is not a valid regular expression", n(name)));
            }
            if (result != null) {
                return result;
            } else {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("the value of %s is not a valid regular expression (null result)", n(name)));
            }
        }
        
        @CheckReturnValue
        public Pattern pattern(String name, Pattern defaultValue) throws OXException {
            final Optional<String> str = optional.str(name);
            if (str.isPresent()) {
                final Pattern result;
                try {
                    result = Pattern.compile(str.get());
                } catch (PatternSyntaxException e) {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("the value of %s is not a valid regular expression", n(name)));
                }
                if (result != null) {
                    return result;
                } else {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("the value of %s is not a valid regular expression (null result)", n(name)));
                }
            } else {
                return defaultValue;
            }
        }
        
        @CheckReturnValue
        public Pattern pattern(String name, int flags, Pattern defaultValue) throws OXException {
            final Optional<String> str = optional.str(name);
            if (str.isPresent()) {
                final Pattern result;
                try {
                    result = Pattern.compile(str.get(), flags);
                } catch (PatternSyntaxException e) {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("the value of %s is not a valid regular expression", n(name)));
                }
                if (result != null) {
                    return result;
                } else {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("the value of %s is not a valid regular expression (null result)", n(name)));
                }
            } else {
                return defaultValue;
            }
        }
        
        @CheckReturnValue
        public Pattern pattern(String name, String defaultValue) throws OXException {
            final Optional<String> str = optional.str(name);
            if (str.isPresent()) {
                final Pattern result;
                try {
                    result = Pattern.compile(str.get());
                } catch (PatternSyntaxException e) {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("the value of %s is not a valid regular expression", n(name)));
                }
                if (result != null) {
                    return result;
                } else {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("the value of %s is not a valid regular expression (null result)", n(name)));
                }
            } else {
                final Pattern result;
                try {
                    result = Pattern.compile(defaultValue);
                } catch (PatternSyntaxException e) {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("the default value for %s is not a valid regular expression", n(name)));
                }
                if (result != null) {
                    return result;
                } else {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("the default value for %s is not a valid regular expression (null result)", n(name)));
                }
            }
        }
        
        @CheckReturnValue
        public Pattern pattern(String name, int flags, String defaultValue) throws OXException {
            final Optional<String> str = optional.str(name);
            if (str.isPresent()) {
                final Pattern result;
                try {
                    result = Pattern.compile(str.get(), flags);
                } catch (PatternSyntaxException e) {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("the value of %s is not a valid regular expression", n(name)));
                }
                if (result != null) {
                    return result;
                } else {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("the value of %s is not a valid regular expression (null result)", n(name)));
                }
            } else {
                final Pattern result;
                try {
                    result = Pattern.compile(defaultValue, flags);
                } catch (PatternSyntaxException e) {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("the default value for %s is not a valid regular expression", n(name)));
                }
                if (result != null) {
                    return result;
                } else {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("the default value for %s is not a valid regular expression (null result)", n(name)));
                }
            }
        }
        
        @CheckReturnValue
        public Pattern pattern(String name, int flags) throws OXException {
            final String str = mandatory.str(name);
            final Pattern result;
            try {
                result = Pattern.compile(str, flags);
            } catch (PatternSyntaxException e) {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("the value of %s is not a valid regular expression", n(name)));
            }
            if (result != null) {
                return result;
            } else {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("the value of %s is not a valid regular expression (null result)", n(name)));
            }
        }
        
    };
    public final MandatoryValueReader mandatory = new MandatoryValueReader();
    public final MandatoryValueReader m = mandatory;

    public final class OptionalValueReader {
        @CheckReturnValue
        public Optional<String> str(final String name) throws OXException {
            final String n = n(name);
            final ConfigView cv = configView;
            if (cv == null) {
                return Optional.absent();
            }
            String str = cv.opt(n, String.class, null);
            if (str == null) {
                return Optional.absent();
            }
            str = str.replaceAll("^\\s*", "").replaceAll("\\s*$", "");
            if (str.isEmpty()) {
                return Optional.absent();
            }
            return Optional.of(str);
        }
        
        @CheckReturnValue
        public Optional<Long> longNum(final String name) throws OXException {
            final Optional<String> str = optional.str(name);
            if (! str.isPresent()) {
                return Optional.absent();
            }
            try {
                return Optional.fromNullable(Long.parseLong(str.get()));
            } catch (final NumberFormatException e) {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("%s is not a numeric value", n(name)));
            }
        }

        @CheckReturnValue
        public Optional<Integer> intNum(final String name) throws OXException {
            final Optional<String> str = optional.str(name);
            if (! str.isPresent()) {
                return Optional.absent();
            }
            try {
                return Optional.fromNullable(Integer.parseInt(str.get()));
            } catch (final NumberFormatException e) {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("%s is not a numeric value", n(name)));
            }
        }
        @CheckReturnValue
        public Optional<Integer> positiveIntNum(final String name) throws OXException {
            final Optional<Integer> o = optional.intNum(name);
            if (o.isPresent()) {
                final int value = o.get();
                if (value < 0) {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("%s must be a positive numeric value", n(name)));
                }
            }
            return o;
        }
        
        @CheckReturnValue
        public Optional<Boolean> bool(final String name) throws OXException {
            final Optional<String> o = optional.str(name);
            if (! o.isPresent()) {
                return Optional.absent();
            }
            final String str = o.get();
            if (str.equalsIgnoreCase("true")) {
                return Optional.of(true);
            }
            if (str.equalsIgnoreCase("false")) {
                return Optional.of(false);
            }
            throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("%s is not a boolean (true/false)", n(name)));
        }
        
        @CheckReturnValue
        public Optional<String> hostname(final String name) throws OXException {
            final Optional<String> o = optional.str(name);
            if (! o.isPresent()) {
                return Optional.absent();
            }
            return Optional.fromNullable(IDNA.toASCII(o.get()));
        }
        
        @CheckReturnValue
        public Optional<Integer> port(final String name) throws OXException {
            final Optional<Integer> i = optional.intNum(name);
            if (! i.isPresent()) {
                return i;
            }
            final int value = i.get();
            if (value <= 0 || value > 65535) {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("%s does not contain a valid port number", n(name)));
            } else {
                return i;
            }
        }
        
        @CheckReturnValue
        public Optional<Charset> charset(final String name) throws OXException {
            final Optional<String> o = optional.str(name);
            if (! o.isPresent()) {
                return Optional.absent();
            } else {
                try {
                    return Optional.fromNullable(Charset.forName(o.get()));
                } catch (final Exception e) {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("%s does not contain a valid charset: %s",
                        n(name), e.getMessage()));
                }
            }
        }
        
        @CheckReturnValue
        public ImmutableList<String> strList(final String name, final CharMatcher split, final Iterator<String> defaultValues) throws OXException {
            final Optional<String> o = optional.str(name);
            if (! o.isPresent()) {
                return ImmutableList.<String>copyOf(defaultValues);
            }
            return ImmutableList.copyOf(Splitter.on(split).omitEmptyStrings().split(o.get()));
        }
        @CheckReturnValue
        public ImmutableList<String> strList(final String name, final Iterator<String> defaultValues) throws OXException {
            return optional.strList(name, DEFAULT_SPLIT_PATTERN, defaultValues);
        }
        @CheckReturnValue
        public ImmutableList<String> strList(final String name, final CharMatcher split, final String... defaultValues) throws OXException {
            return optional.strList(name, split, Iterators.forArray(defaultValues));
        }
        @CheckReturnValue
        public ImmutableList<String> strList(final String name, final String... defaultValues) throws OXException {
            return optional.strList(name, DEFAULT_SPLIT_PATTERN, Iterators.forArray(defaultValues));
        }
        
        @CheckReturnValue
        public ImmutableList<String> strList(final String name, final CharMatcher split) throws OXException {
            final Optional<String> o = optional.str(name);
            if (! o.isPresent()) {
                return ImmutableList.of();
            }
            return ImmutableList.copyOf(Splitter.on(split).omitEmptyStrings().split(o.get()));
        }
        @CheckReturnValue
        public ImmutableList<String> strList(final String name) throws OXException {
            return optional.strList(name, DEFAULT_SPLIT_PATTERN);
        }
        
        @CheckReturnValue
        public ImmutableList<String> strList(final String name, final CharMatcher split, final Iterable<String> defaultValues) throws OXException {
            final Optional<String> o = optional.str(name);
            if (! o.isPresent()) {
                return ImmutableList.<String>copyOf(defaultValues);
            }
            return ImmutableList.copyOf(Splitter.on(split).omitEmptyStrings().split(o.get()));
        }
        @CheckReturnValue
        public ImmutableList<String> strList(final String name, final Iterable<String> defaultValues) throws OXException {
            return optional.strList(name, DEFAULT_SPLIT_PATTERN, defaultValues);
        }
        
        @CheckReturnValue
        public ImmutableSet<String> strSet(final String name, final CharMatcher split, final Iterator<String> defaultValues) throws OXException {
            final Optional<String> o = optional.str(name);
            if (! o.isPresent()) {
                return ImmutableSet.<String>copyOf(defaultValues);
            }
            return ImmutableSet.copyOf(Splitter.on(split).omitEmptyStrings().split(o.get()));
        }
        @CheckReturnValue
        public ImmutableSet<String> strSet(final String name, final Iterator<String> defaultValues) throws OXException {
            return optional.strSet(name, DEFAULT_SPLIT_PATTERN, defaultValues);
        }
        @CheckReturnValue
        public ImmutableSet<String> strSet(final String name, final CharMatcher split, final String... defaultValues) throws OXException {
            return optional.strSet(name, split, Iterators.forArray(defaultValues));
        }
        @CheckReturnValue
        public ImmutableSet<String> strSet(final String name, final String... defaultValues) throws OXException {
            return optional.strSet(name, DEFAULT_SPLIT_PATTERN, Iterators.forArray(defaultValues));
        }
        
        @CheckReturnValue
        public ImmutableSet<String> strSet(final String name, final CharMatcher split) throws OXException {
            final Optional<String> o = optional.str(name);
            if (! o.isPresent()) {
                return ImmutableSet.of();
            }
            return ImmutableSet.copyOf(Splitter.on(split).omitEmptyStrings().split(o.get()));
        }
        @CheckReturnValue
        public ImmutableSet<String> strSet(final String name) throws OXException {
            return optional.strSet(name, DEFAULT_SPLIT_PATTERN);
        }
        
        @CheckReturnValue
        public ImmutableSet<String> strSet(final String name, final CharMatcher split, final Iterable<String> defaultValues) throws OXException {
            final Optional<String> o = optional.str(name);
            if (! o.isPresent()) {
                return ImmutableSet.<String>copyOf(defaultValues);
            }
            return ImmutableSet.copyOf(Splitter.on(split).omitEmptyStrings().split(o.get()));
        }
        @CheckReturnValue
        public ImmutableSet<String> strSet(final String name, final Iterable<String> defaultValues) throws OXException {
            return optional.strSet(name, DEFAULT_SPLIT_PATTERN, defaultValues);
        }
        
        @CheckReturnValue
        public Optional<URI> uri(String name) throws OXException {
            final Optional<String> o = optional.str(name);
            if (! o.isPresent()) {
                return Optional.absent();
            }
            try {
                return Optional.of(new URI(o.get()));
            } catch (final URISyntaxException e) {
                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("the value of %s is not a valid URI", n(name)));
            }
        }
        
        @CheckReturnValue
        public Optional<Pattern> pattern(String name) throws OXException {
            final Optional<String> str = optional.str(name);
            if (str.isPresent()) {
                final Pattern result;
                try {
                    result = Pattern.compile(str.get());
                } catch (PatternSyntaxException e) {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("the value of %s is not a valid regular expression", n(name)));
                }
                if (result != null) {
                    return Optional.of(result);
                } else {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("the value of %s is not a valid regular expression (null result)", n(name)));
                }
            } else {
                return Optional.absent();
            }
        }
        
        @CheckReturnValue
        public Optional<Pattern> pattern(String name, int flags) throws OXException {
            final Optional<String> str = optional.str(name);
            if (str.isPresent()) {
                final Pattern result;
                try {
                    result = Pattern.compile(str.get(), flags);
                } catch (PatternSyntaxException e) {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, String.format("the value of %s is not a valid regular expression", n(name)));
                }
                if (result != null) {
                    return Optional.of(result);
                } else {
                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(String.format("the value of %s is not a valid regular expression (null result)", n(name)));
                }
            } else {
                return Optional.absent();
            }
        }
    };
    public final OptionalValueReader optional = new OptionalValueReader();
    public final OptionalValueReader o = optional;

    public static final CharMatcher DEFAULT_SPLIT_PATTERN = CharMatcher.anyOf("\r\n;, ");
    
}
